/*
 * Copyright 2024 Luca Fulchir <luca.fulchir@runesauth.com>
 *
 * Licensed under the Apache License, Version 2.0 with LLVM exception (the
 * "License"); you may not use this file except in compliance with the
 * License. You may obtain a copy of the License and of the exception at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     https://spdx.org/licenses/LLVM-exception.html
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

use ::std::{
    collections::BTreeMap,
    sync::{Arc, RwLock},
    vec::Vec,
};

use crate::{Path, PkgBuilder, PkgName, RepoName, Repository};

#[derive(Debug, ::thiserror::Error)]
pub enum Error {
    #[error("package already present")]
    AlreadyPresent((Arc<dyn Repository>, ArcLock<dyn PkgBuilder>)),
    #[error("not found")]
    NotFound,
}

type ArcLock<T> = Arc<RwLock<T>>;
type PkgMap =
    BTreeMap<Path<PkgName>, (Arc<dyn Repository>, ArcLock<dyn PkgBuilder>)>;

/// Collection of packages, subset of what is present on one repository only
pub struct Collection {
    repos: RwLock<Vec<Arc<dyn Repository>>>,
    repos_pkgs: RwLock<BTreeMap<Path<RepoName>, PkgMap>>,
}

impl Collection {
    /// Create a new empty collection of pkgs
    pub fn new() -> Collection {
        Collection {
            repos: RwLock::new(Vec::with_capacity(8)),
            repos_pkgs: RwLock::new(BTreeMap::new()),
        }
    }
    pub fn add_pkg(
        &self,
        r: Arc<dyn Repository>,
        p: ArcLock<dyn PkgBuilder>,
    ) -> Result<(), Error> {
        let repo_name = r.path();
        let mut repos = self.repos.write().unwrap();
        let mut repos_pkgs = self.repos_pkgs.write().unwrap();
        if repos
            .binary_search_by(|x| x.path().cmp(&repo_name))
            .is_err()
        {
            repos.push(r.clone());
            repos.sort_by(|a, b| a.path().cmp(&b.path()))
        }
        let pkg_name = p.read().unwrap().path();
        match repos_pkgs.get_mut(&repo_name) {
            Some(pkg_map) => match pkg_map.get(&pkg_name) {
                Some((r, p)) => {
                    return Err(Error::AlreadyPresent((r.clone(), p.clone())));
                }
                None => {
                    pkg_map.insert(pkg_name, (r, p));
                }
            },
            None => {
                let mut pkg_map: PkgMap = BTreeMap::new();
                pkg_map.insert(pkg_name, (r, p));
                repos_pkgs.insert(repo_name, pkg_map);
            }
        }

        Ok(())
    }
    pub fn find(
        &self,
        pkg_path: Path<PkgName>,
    ) -> Result<(Arc<dyn Repository>, ArcLock<dyn PkgBuilder>), Error> {
        let repos_pkgs = self.repos_pkgs.read().unwrap();
        for (_, pkg_map) in repos_pkgs.iter() {
            match pkg_map.get(&pkg_path) {
                Some((r, p)) => {
                    return Ok((r.clone(), p.clone()));
                }
                None => {}
            }
        }
        Err(Error::NotFound)
    }
    pub fn find_in_repo(
        &self,
        repo_path: Path<RepoName>,
        pkg_path: Path<PkgName>,
    ) -> Result<(Arc<dyn Repository>, ArcLock<dyn PkgBuilder>), Error> {
        let repos_pkgs = self.repos_pkgs.read().unwrap();
        match repos_pkgs.get(&repo_path) {
            None => Err(Error::NotFound),
            Some(pkg_map) => match pkg_map.get(&pkg_path) {
                Some((r, p)) => Ok((r.clone(), p.clone())),
                None => Err(Error::NotFound),
            },
        }
    }
}
