/*

* Copyright 2024 Luca Fulchir <luca.fulchir@runesauth.com>
*
* Licensed under the Apache License, Version 2.0 with LLVM exception (the
* "License"); you may not use this file except in compliance with the
* License. You may obtain a copy of the License and of the exception at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*     https://spdx.org/licenses/LLVM-exception.html
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

use ::std::sync::Arc;

mod collection;
mod names;
pub use collection::Collection;
pub use names::{Path, PkgName, RepoName};
pub mod deps {
    pub struct Build(pub crate::Collection);
    pub struct Runtime(pub crate::Collection);
}

pub use ::bok_macro::{
    impl_package, package, pkg_fn_to_code, repository, Package, Repository,
};
pub use ::semver::{BuildMetadata, Prerelease, Version};

//use ::std::any::Any;

// Package stuff

/// Get a package from its module and export it in the current module
#[macro_export]
macro_rules! moduse {
    ($($name:ident,)*) => {
        $(
            pub mod $name;
            #[allow(unused_imports)]
            pub use $name::*;
        )*
    };
}
// re-export `paste` crate for next macros
pub use ::paste;

/// Add multipla packages to a repo
/// e.g.:
/// bok::repo_packages! {
///   MyRepo {
///     Mypkg1,
///     Mypkg2,
///   }
/// }
#[macro_export]
macro_rules! repo_packages {
    ( $repo:ident { $($name:ident,)*} )  => {
        impl $repo {
            $crate::packages!{$($name ,)*}
        }
    };
}

/// Add multipla packages to a repo Impl
/// e.g.:
/// impl MyRepo {
///   bok::packages! {
///     Mypkg1,
///     Mypkg2,
///   }
/// }
#[macro_export]
macro_rules! packages {
    ($($name:ident,)*) => {
        $crate::paste::paste! {
            $(
                pub fn [<$name:snake>] (&self) -> $name {
                    $name::default()
                }
                pub fn [<$name:snake _builder>] (&self) -> [<$name Builder>] {
                    $name::builder()
                }
            )*
        }
    };
}

/// Marks your struct as a repository
pub trait Repository: ::core::fmt::Debug {
    fn name(&self) -> RepoName;
    fn path(&self) -> Path<RepoName>;
}

/// This is an empty Package List.
#[derive(::std::default::Default, Debug)]
pub struct RepositoryEmpty {}
impl Repository for RepositoryEmpty {
    fn name(&self) -> RepoName {
        "RepositoryEmpty".into()
    }
    fn path(&self) -> Path<RepoName> {
        "::bok::RepositoryEmpty".into()
    }
}

/// Print the code of the package
pub trait PkgCode {
    fn fmt(&self, f: &mut ::core::fmt::Formatter<'_>) -> ::core::fmt::Result;
}

/// Implement common package operations
pub trait Pkg:
    PkgCode + ::core::fmt::Debug + ::core::fmt::Display + ::std::any::Any
{
    fn name(&self) -> PkgName;
    fn path(&self) -> Path<PkgName>;
    fn version(&self) -> Version;
    fn base(&self) -> ::core::option::Option<&dyn Pkg>;
    fn base_mut(&mut self) -> ::core::option::Option<&mut dyn Pkg>;
    fn as_any(&self) -> &dyn ::std::any::Any;
    fn as_any_mut(&mut self) -> &mut dyn ::std::any::Any;
    fn dependencies_set(
        &self,
        repo: Arc<dyn Repository>,
        build: &deps::Build,
        runtime: &deps::Runtime,
    ) -> Result<(), ()>;
    fn prepare(&self) -> ::core::result::Result<(), ()> {
        self.base().unwrap().prepare()
    }
    fn prepare_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            self.base().unwrap().prepare()
        }
    }
    fn configure(&self) -> ::core::result::Result<(), ()> {
        self.base().unwrap().configure()
    }
    fn configure_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            self.base().unwrap().configure()
        }
    }
    fn build(&self) -> ::core::result::Result<(), ()> {
        self.base().unwrap().build()
    }
    fn build_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            self.base().unwrap().build()
        }
    }
    fn check(&self) -> ::core::result::Result<(), ()> {
        self.base().unwrap().check()
    }
    fn check_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            self.base().unwrap().check()
        }
    }
    fn install(&self) -> ::core::result::Result<(), ()> {
        self.base().unwrap().install()
    }
    fn install_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            self.base().unwrap().install()
        }
    }
    fn package(&self) -> ::core::result::Result<(), ()> {
        self.prepare()?;
        self.configure()?;
        self.build()?;
        self.check()?;
        self.install()
    }
    fn hash(&self) -> Hash;
    fn hash_code(&self) -> Hash;
}

pub trait PkgBuilder: ::core::fmt::Debug + ::std::any::Any {
    fn name(&self) -> crate::PkgName;
    fn path(&self) -> crate::Path<crate::PkgName>;
    fn version(&self) -> crate::Version;
    fn default_unused(&mut self) -> &mut dyn crate::PkgBuilder;
    fn build(
        &mut self,
    ) -> Result<Box<dyn Pkg>, ::std::boxed::Box<dyn ::std::error::Error>>;
}

#[derive(Debug, Clone)]
pub struct PkgEmpty {
    version: Version,
}

impl Default for PkgEmpty {
    fn default() -> Self {
        PkgEmpty {
            version: Version {
                major: 0,
                minor: 0,
                patch: 0,
                pre: Prerelease::EMPTY,
                build: BuildMetadata::EMPTY,
            },
        }
    }
}

impl ::core::fmt::Display for PkgEmpty {
    fn fmt(&self, _f: &mut ::core::fmt::Formatter<'_>) -> ::core::fmt::Result {
        Ok(())
    }
}

impl PkgCode for PkgEmpty {
    fn fmt(&self, _f: &mut ::core::fmt::Formatter<'_>) -> ::core::fmt::Result {
        Ok(())
    }
}

impl Pkg for PkgEmpty {
    fn name(&self) -> PkgName {
        "PkgEmpty".into()
    }
    fn path(&self) -> Path<PkgName> {
        "::bok::Pkg::PkgEmpty".into()
    }
    fn version(&self) -> Version {
        self.version.clone()
    }
    fn base(&self) -> Option<&dyn Pkg> {
        None
    }
    fn base_mut(&mut self) -> Option<&mut dyn Pkg> {
        None
    }
    fn as_any(&self) -> &dyn ::std::any::Any {
        self
    }
    fn as_any_mut(&mut self) -> &mut dyn ::std::any::Any {
        self
    }
    fn dependencies_set(
        &self,
        _repo: Arc<dyn Repository>,
        _build: &deps::Build,
        _runtime: &deps::Runtime,
    ) -> Result<(), ()> {
        Ok(())
    }
    fn prepare(&self) -> ::core::result::Result<(), ()> {
        ::core::result::Result::Ok(())
    }
    fn prepare_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            ret = ::core::result::Result::Ok(());
        }
    }
    fn configure(&self) -> ::core::result::Result<(), ()> {
        ::core::result::Result::Ok(())
    }
    fn configure_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            ret = ::core::result::Result::Ok(());
        }
    }
    fn build(&self) -> ::core::result::Result<(), ()> {
        ::core::result::Result::Ok(())
    }
    fn build_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            ret = ::core::result::Result::Ok(());
        }
    }
    fn check(&self) -> ::core::result::Result<(), ()> {
        ::core::result::Result::Ok(())
    }
    fn check_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            ret = ::core::result::Result::Ok(());
        }
    }
    fn install(&self) -> ::core::result::Result<(), ()> {
        ::core::result::Result::Ok(())
    }
    fn install_code(&self) -> ::proc_macro2::TokenStream {
        ::quote::quote! {
            ret = ::core::result::Result::Ok(());
        }
    }
    fn hash(&self) -> Hash {
        use ::sha3::Digest;
        let hasher = ::sha3::Sha3_256::new();
        let hash = hasher.finalize();
        Hash::Sha3(hash)
    }
    fn hash_code(&self) -> Hash {
        use ::sha3::Digest;
        let hasher = ::sha3::Sha3_256::new();
        let hash = hasher.finalize();
        Hash::Sha3(hash)
    }
}

::bitflags::bitflags! {
    /// In each package, what we can build and how
    #[repr(transparent)]
    #[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
    pub struct Build: u64 {
        /// only the binary
        const Binary = 0b1;
        /// only the library
        const Library = 0b10;
        /// only the development headers
        const Development = 0b100;
        /// only the documentation
        const Documentation = 0b1000;
        /// Run checks during build
        const Checks = 0b10000;
        /// Strip binaries after build
        const Strip = 0b100000;
        /// Build in release mode
        const Release = 0b1000000;
        /// Build in debug mode
        const Debug = 0b10000000;
        /// Build with profiling
        const Profile = 0b10000000;
        /// Build optimised for size
        const MinSize = 0b100000000;
    }
}

// Conf stuff
pub struct Conf {}

/// package hash
#[derive(Debug, Copy, Clone)]
pub enum Hash {
    Sha3(::sha3::digest::Output<::sha3::Sha3_256>),
}

impl ::std::fmt::Display for Hash {
    fn fmt(&self, f: &mut ::std::fmt::Formatter<'_>) -> ::std::fmt::Result {
        match self {
            //FIXME: move to something like zbase85
            Hash::Sha3(h) => write!(f, "{:x}", h),
        }
    }
}

/*
pub enum Value<T: Sized> {
    /// Set, only once per
    Set(T),
    /// Delete a value
    Del(T),
}
pub enum ValueList<T: Sized> {
    /// Set, only once per
    Set(T),
    /// Delete a value
    Del(T),
    /// Add to a list
    Add(T),
    /// Merge (add + dedup)
    Merge(T),
}
*/
