/*

* Copyright 2024 Luca Fulchir <luca.fulchir@runesauth.com>
*
* Licensed under the Apache License, Version 2.0 with LLVM exception (the
* "License"); you may not use this file except in compliance with the
* License. You may obtain a copy of the License and of the exception at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*     https://spdx.org/licenses/LLVM-exception.html
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

/// Strong typedef for a repository name
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone)]
pub struct RepoName(pub String);
impl ::std::fmt::Display for RepoName {
    fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {
        write!(f, "{}", self.0)
    }
}
impl From<String> for RepoName {
    fn from(raw: String) -> RepoName {
        RepoName(raw)
    }
}
impl From<&str> for RepoName {
    fn from(raw: &str) -> RepoName {
        RepoName(raw.to_owned())
    }
}

#[derive(PartialEq, Eq, PartialOrd, Ord, Hash, Clone)]
pub struct Path<T> {
    pub path: Vec<String>,
    pub name: T,
}
impl<'a, T> Path<T>
where
    T: From<&'a str>,
{
    pub fn new(raw: &'a str) -> Path<T> {
        Path {
            path: Vec::new(),
            name: raw.into(),
        }
    }
}
impl<'a, T> From<&'a str> for Path<T>
where
    T: From<&'a str>,
{
    fn from(raw: &'a str) -> Self {
        Path::new(raw)
    }
}

impl<T> ::std::fmt::Debug for Path<T>
where
    T: ::std::fmt::Display + ::std::fmt::Debug,
{
    fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {
        let mut path = "".to_owned();
        for p in self.path.iter() {
            path += "::";
            path += p.as_str();
        }
        write!(f, "{}::{}", path, self.name)
    }
}
impl<T> ::std::fmt::Display for Path<T>
where
    T: ::std::fmt::Display + ::std::fmt::Debug,
{
    fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {
        write!(f, "{:?}", self)
    }
}

/// Strong typedef for a package name
#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Hash, Clone)]
pub struct PkgName(pub String);
impl ::std::fmt::Display for PkgName {
    fn fmt(&self, f: &mut ::std::fmt::Formatter) -> ::std::fmt::Result {
        write!(f, "{}", self.0)
    }
}
impl From<String> for PkgName {
    fn from(raw: String) -> PkgName {
        PkgName(raw)
    }
}
impl From<&str> for PkgName {
    fn from(raw: &str) -> PkgName {
        PkgName(raw.to_owned())
    }
}
